@@ -2,20 +2,15 @@ package App::SimpleBackuper;
 
 use strict;
 use warnings;
-use feature ':5.14';
+use feature ':5.'.substr($], 3, 2);
 use Carp;
 use Try::Tiny;
 use Time::HiRes qw(time);
-use Const::Fast;
 use App::SimpleBackuper::BackupDB;
 use App::SimpleBackuper::_format;
 use App::SimpleBackuper::_BlockDelete;
 use App::SimpleBackuper::_BlocksInfo;
 
-const my $SIZE_OF_TOP_FILES => 10;
-const my $SAVE_DB_PERIOD => 60 * 60;
-const my $PRINT_PROGRESS_PERIOD => 60;
-
 sub _proc_uid_gid($$$) {
 	my($uid, $gid, $uids_gids) = @_;
 	
@@ -42,36 +37,6 @@ sub _proc_uid_gid($$$) {
 	return $uid, $gid;
 }
 
-sub _get_block_to_delete {
-	my($state) = @_;
-	
-	if(ref($state->{blocks2delete_prio2size2chunks}) eq 'HASH') {
-		$state->{blocks2delete_prio2size2chunks} = [
-			map {$state->{blocks2delete_prio2size2chunks}->{ $_ }}
-			sort {$a <=> $b}
-			keys %{ $state->{blocks2delete_prio2size2chunks} }
-		];
-	}
-	
-	return if ! @{ $state->{blocks2delete_prio2size2chunks} };
-	
-	if(ref($state->{blocks2delete_prio2size2chunks}->[0]) eq 'HASH') {
-		$state->{blocks2delete_prio2size2chunks}->[0] = [
-			map {$state->{blocks2delete_prio2size2chunks}->[0]->{ $_ }}
-			sort {$b <=> $a}
-			keys %{ $state->{blocks2delete_prio2size2chunks}->[0] }
-		];
-	}
-	
-	my $prio_basket = $state->{blocks2delete_prio2size2chunks}->[0];
-	my $size_basket = $prio_basket->[0];
-	my $block_id = shift @$size_basket;
-	shift @$prio_basket if ! @$size_basket;
-	shift @{ $state->{blocks2delete_prio2size2chunks} } if ! @$prio_basket;
-	
-	return $block_id;
-}
-
 sub Backup {
 	my($options, $state) = @_;
 	
@@ -82,18 +47,17 @@ sub Backup {
 	$state->{$_} = 0 foreach qw(last_backup_id last_file_id last_block_id bytes_processed bytes_in_last_backup total_weight);
 	
 	print "Preparing to backup: " if $options->{verbose};
-	$state->{profile}->{init_ids} = - time();
+	$state->{profile}->{init_ids} = -time;
 	foreach (@$backups) {
 		my $id = $backups->unpack($_)->{id};
 		$state->{last_backup_id} = $id if ! $state->{last_backup_id} or $state->{last_backup_id} < $id;
+		
 	}
 	#print "last backup id $state->{last_backup_id}, ";
 	foreach (@$files) {
 		my $file = $files->unpack($_);
 		$state->{last_file_id} = $file->{id} if ! $state->{last_file_id} or $state->{last_file_id} < $file->{id};
-		if($file->{versions} and @{ $file->{versions} } and $file->{versions}->[-1]->{backup_id_max} == $state->{last_backup_id}) {
-			$state->{bytes_in_last_backup} += $file->{versions}->[-1]->{size};
-		}
+		$state->{bytes_in_last_backup} += $file->{versions}->[-1]->{size} if $file->{versions} and @{ $file->{versions} };
 	}
 	#print "last file id $state->{last_file_id}, ";
 	foreach (@$blocks) {
@@ -104,31 +68,28 @@ sub Backup {
 	$state->{profile}->{init_ids} += time;
 	
 	print "total weight " if $options->{verbose};
-	for(my $q = 0; $q <= $#$parts; $q++) {
-		$state->{total_weight} += $parts->unpack($parts->[ $q ])->{size};
-	}
+	$state->{total_weight} += $parts->unpack($_)->{size} foreach @$parts;
 	print fmt_weight($state->{total_weight}).", " if $options->{verbose};
 	
 	my $cur_backup = {name => $options->{'backup-name'}, id => ++$state->{last_backup_id}, files_cnt => 0, max_files_cnt => 0};
 	$backups->upsert({ id => $cur_backup->{id} }, $cur_backup);
 	
 	{
-		$state->{blocks_info} = _BlocksInfo($options, $state);
-		
-		$state->{blocks2delete_prio2size2chunks} = {};
-		foreach my $block (@$blocks) {
-			my $block_id = $blocks->unpack($block)->{id};
-			next if ! $block_id; # What's this?
-			my $block_info = $state->{blocks_info}->{ $block_id };
-			push @{	$state->{blocks2delete_prio2size2chunks}->{ $block_info->[0] }->{ $block_info->[1] } }, $block_id;
-		}
-		
+		#print "blocks stack to delete...";
+		my $blocks_info = _BlocksInfo($options, $state);
+		$state->{blocks_stack2delete} = [
+			map {[ $_, @{ $blocks_info->{$_}->[2] } ]} sort {
+				$blocks_info->{$a}->[0] <=> $blocks_info->{$b}->[0]
+				or $blocks_info->{$b}->[1] <=> $blocks_info->{$a}->[1]
+			} grep {$_} map { $blocks->unpack( $_ )->{id} } @$blocks
+		];
 		print " OK\n" if $options->{verbose};
 	}
 	
 	_print_progress($state) if $options->{verbose};
 	
-	my(%files_queues_by_priority, %dirs2upd);
+	
+	my %files_queues_by_priority;
 	while(my($mask, $priority) = each %{ $options->{files} }) {
 		next if ! $priority;
 		foreach my $path (glob $mask) {
@@ -152,31 +113,30 @@ sub Backup {
 					push @cur_path, $path_node;
 					
 					my $file = $files->find_by_parent_id_name($file_id, $path_node);
-					$file //= {
-						parent_id	=> $file_id,
-						id			=> ++$state->{last_file_id},
-						name		=> $path_node,
-						versions	=> [ {
-							backup_id_min	=> $state->{last_backup_id},
-							backup_id_max	=> 0,
-							uid				=> 0,
-							gid				=> 0,
-							size			=> 0,
-							mode			=> 0,
-							mtime			=> 0,
-							block_id		=> 0,
-							symlink_to		=> undef,
-							parts			=> [],
-						} ],
-					};
-					
-					$dirs2upd{join('/', @cur_path) || '/'} = {
-						parent_id	=> $file_id,
-						filename	=> $path_node,
-					};
-						
-					$files->upsert({ id => $file->{id}, parent_id => $file->{parent_id} }, $file);
-					
+					if(! $file) {
+						my @stat = lstat(join('/', @cur_path) || '/');
+						my($uid, $gid) =_proc_uid_gid($stat[4], $stat[5], $state->{db}->{uids_gids});
+						$file = {
+							parent_id	=> $file_id,
+							id			=> ++$state->{last_file_id},
+							name		=> $path_node,
+							versions	=> [
+								{
+									backup_id_min	=> $state->{last_backup_id},
+									backup_id_max	=> $state->{last_backup_id},
+									uid				=> $uid,
+									gid				=> $gid,
+									size			=> $stat[7],
+									mode			=> $stat[2],
+									mtime			=> $stat[9],
+									block_id		=> 0,
+									symlink_to		=> undef,
+									parts			=> [],
+								}
+							],
+						};
+						$files->upsert({ id => $file->{id}, parent_id => $file->{parent_id} }, $file);
+					}
 					$file_id = $file->{id};
 				}
 			}
@@ -186,85 +146,26 @@ sub Backup {
 	}
 	delete $files_queues_by_priority{ $_ } foreach grep {! @{ $files_queues_by_priority{ $_ } }} keys %files_queues_by_priority;
 	
-	my $last_db_save = time;
-	my $last_print_progress = time;
 	while(%files_queues_by_priority) {
 		my($priority) = sort {$b <=> $a} keys %files_queues_by_priority;
 		my $task = shift @{ $files_queues_by_priority{$priority} };
 		delete $files_queues_by_priority{$priority} if ! @{ $files_queues_by_priority{$priority} };
 		my @next = _file_proc( $task, $options, $state );
 		unshift @{ $files_queues_by_priority{ $_->[1] } }, $_ foreach reverse @next;
-		
-		if($options->{verbose} and time - $last_print_progress > $PRINT_PROGRESS_PERIOD) {
-			_print_progress($state);
-			$last_print_progress = time;
-		}
-		
-		if(time - $last_db_save > $SAVE_DB_PERIOD) {
-			App::SimpleBackuper::BackupDB($options, $state);
-			$last_db_save = time;
-		}
 	}
 	
-	while(my($full_path, $dir2upd) = each %dirs2upd) {
-		print "Updating dir $full_path..." if $options->{verbose};
-		my $file = $files->find_by_parent_id_name($dir2upd->{parent_id}, $dir2upd->{filename});
-		my @stat = lstat($full_path);
-		if(@stat and $file->{versions}->[-1]->{backup_id_max} != $state->{last_backup_id}) {
-			my($uid, $gid) =_proc_uid_gid($stat[4], $stat[5], $state->{db}->{uids_gids});
-			if($file->{versions}->[-1]->{backup_id_max} == $state->{last_backup_id} - 1) {
-				$file->{versions}->[-1] = {
-					%{ $file->{versions}->[-1] },
-					backup_id_max	=> $state->{last_backup_id},
-					uid				=> $uid,
-					gid				=> $gid,
-					size			=> $stat[7],
-					mode			=> $stat[2],
-					mtime			=> $stat[9],
-					block_id		=> 0,
-					symlink_to		=> undef,
-					parts			=> [],
-				};
-			} else {
-				push @{ $file->{versions} }, {
-					backup_id_min	=> $state->{last_backup_id},
-					backup_id_max	=> $state->{last_backup_id},
-					uid				=> $uid,
-					gid				=> $gid,
-					size			=> $stat[7],
-					mode			=> $stat[2],
-					mtime			=> $stat[9],
-					block_id		=> 0,
-					symlink_to		=> undef,
-					parts			=> [],
-				}
-			}
-			$files->upsert({ id => $file->{id}, parent_id => $file->{parent_id} }, $file);
-			
-			my $backup = $backups->find_row({ id => $state->{last_backup_id} });
-			$backup->{files_cnt}++;
-			$backup->{max_files_cnt}++;
-			$backups->upsert({ id => $backup->{id} }, $backup );
-		}
-		
-		print "OK\n" if $options->{verbose};
-	}
-	
-	my $backup = $backups->find_row({ id => $state->{last_backup_id} });
-	$backup->{is_done} = 1;
-	$backups->upsert({ id => $backup->{id} }, $backup );
-	
 	App::SimpleBackuper::BackupDB($options, $state);
 	
-	_print_progress($state) if ! $options->{quiet};
+	_print_progress($state, 1) if ! $options->{quiet};
 }
 
 sub _print_progress {
-	print "Progress: ";
-	if($_[0]->{bytes_in_last_backup}) {
-		printf "processed %s of %s in last backup, ", fmt_weight($_[0]->{bytes_processed}), fmt_weight($_[0]->{bytes_in_last_backup});
-	}
-	printf "total backups weight %s.\n", fmt_weight($_[0]->{total_weight});
+	state $last_print_time = 0;
+	return if time - $last_print_time < 60 and ! $_[1];
+	
+	printf "Progress: processed %s of %s in last backup, total backups weight %s.\n",
+		fmt_weight($_[0]->{bytes_processed}), fmt_weight($_[0]->{bytes_in_last_backup}), fmt_weight($_[0]->{total_weight});
+	$last_print_time = time;
 }
 
 use Text::Glob qw(match_glob);
@@ -274,6 +175,8 @@ use App::SimpleBackuper::RegularFile;
 sub _file_proc {
 	my($task, $options, $state) = @_;
 	
+	_print_progress($state) if $options->{verbose};
+	
 	confess "No task" if ! $task;
 	confess "No filepath" if ! $task->[0];
 	
@@ -325,7 +228,7 @@ sub _file_proc {
 		if($file) {
 			print ", is old file #$file->{id}" if $options->{verbose};
 			if($file->{versions}->[-1]->{backup_id_max} == $state->{last_backup_id}) {
-				print ", is already backuped.\n" if $options->{verbose};
+				print ", is already backuped.\n";
 				return;
 			}
 		} else {
@@ -339,8 +242,6 @@ sub _file_proc {
 		}
 	}
 	
-	$state->{bytes_processed} += $file->{versions}->[-1]->{size} if @{ $file->{versions} };
-	
 	my %version = (
 		backup_id_min	=> $state->{last_backup_id},
 		backup_id_max	=> $state->{last_backup_id},
@@ -354,6 +255,8 @@ sub _file_proc {
 		parts			=> [],
 	);
 	
+	$state->{bytes_processed} += $version{size};
+	
 	if(S_ISDIR $stat[2]) {
 		print ", is directory.\n" if $options->{verbose};
 		my $dh;
@@ -421,7 +324,6 @@ sub _file_proc {
 		
 		if(@{ $file->{versions} } and $file->{versions}->[-1]->{mtime} == $version{mtime}) {
 			$version{parts} = $file->{versions}->[-1]->{parts}; # If mtime not changed then file not changed
-			
 			$version{block_id} = $file->{versions}->[-1]->{block_id};
 			
 			my $block = $blocks->find_row({ id => $version{block_id} });
@@ -440,18 +342,16 @@ sub _file_proc {
 			while(1) {
 				$state->{profile}->{fs} -= time;
 				$state->{profile}->{fs_read} -= time;
-				my $read_failed;
 				my $read = try {
 					$reg_file->read($part_number);
 				} catch {
 					1 while chomp;
 					push @{ $state->{fails}->{$_} }, $task->[0];
 					print ", can't read: $_\n" if $options->{verbose};
-					$read_failed = 1;
 				};
 				$state->{profile}->{fs} += time;
 				$state->{profile}->{fs_read} += time;
-				return if $read_failed;
+				return if ! defined $read;
 				last if ! $read;
 				
 				print "\tpart #$part_number: " if $options->{verbose};
@@ -473,9 +373,7 @@ sub _file_proc {
 				
 				# Search for part with this hash
 				if(my $part = $parts->find_row({ hash => $part{hash} })) {
-					if($part->{block_id}) {
-						$block_ids{ $part->{block_id} }++;
-					}
+					$block_ids{ $part->{block_id} }++ if $part->{block_id};
 					$part{size} = $part->{size};
 					$part{aes_key} = $part->{aes_key};
 					$part{aes_iv} = $part->{aes_iv};
@@ -507,12 +405,13 @@ sub _file_proc {
 					$file_weight_spent += $reg_file->size;
 					
 					if($state->{total_weight} > $options->{space_limit}) {
-						print ", freeing up space by ".fmt_weight($state->{total_weight} - $options->{space_limit})."...\n" if $options->{verbose};
+						print "freeing up space by ".fmt_weight($state->{total_weight} - $options->{space_limit})."\n" if $options->{verbose};
 						$file_time_spent += time;
 						while($state->{total_weight} > $options->{space_limit}) {
 							_free_up_space($options, $state, \%block_ids);
 						}
 						$file_time_spent -= time;
+						print "\t... " if $options->{verbose};
 					}
 					
 					$state->{profile}->{storage} -= time;
@@ -560,12 +459,6 @@ sub _file_proc {
 							next if $version->{block_id} != $bi;
 							$version->{block_id} = $block->{id};
 							$block->{parts_cnt} += @{ $version->{parts} };
-							foreach my $vpart (@{ $version->{parts} }) {
-								my $part = $parts->find_row({ hash => $vpart->{hash} });
-								next if $part->{block_id} == $block->{id};
-								$part->{block_id} = $block->{id};
-								$parts->upsert({ hash => $part->{hash} }, $part);
-							}
 						}
 						$files->[ $block_file_index ] = $files->pack( $block_file );
 					}
@@ -608,13 +501,7 @@ sub _file_proc {
 		and $file->{versions}->[-1]->{size}	== $version{size}
 		and $file->{versions}->[-1]->{mode}	== $version{mode}
 		and $file->{versions}->[-1]->{mtime}== $version{mtime}
-		and (
-			defined $file->{versions}->[-1]->{symlink_to} == defined $version{symlink_to}
-			and (
-				! defined $version{symlink_to}
-				or $file->{versions}->[-1]->{symlink_to} eq $version{symlink_to}
-			)
-		)
+		and ( defined $file->{versions}->[-1]->{symlink_to} == defined $version{symlink_to} or ( defined $version{symlink_to} and $file->{versions}->[-1]->{symlink_to} eq $version{symlink_to} ) )
 		and join(' ', map { $_->{hash} } @{ $file->{versions}->[-1]->{parts} }) eq join(' ', map { $_->{hash} } @{ $version{parts} })
 	) {
 		$file->{versions}->[-1]->{backup_id_max} = $state->{last_backup_id};
@@ -631,23 +518,17 @@ sub _file_proc {
 
 	
 	$state->{longest_files} ||= [];
-	if(	@{ $state->{longest_files} } < $SIZE_OF_TOP_FILES
-		or $state->{longest_files}->[-1]->{time} < $file_time_spent
-	) {
+	if(@{ $state->{longest_files} } < 5 or $state->{longest_files}->[-1]->{time} < $file_time_spent) {
 		@{ $state->{longest_files} } = sort {$b->{time} <=> $a->{time}} (@{ $state->{longest_files} }, {time => $file_time_spent, path => $task->[0]});
-		splice @{ $state->{longest_files} }, $SIZE_OF_TOP_FILES;
+		splice @{ $state->{longest_files} }, 5;
 	}
 	
-	if($file_weight_spent) {
-		$state->{heaviweightest_files} ||= [];
-		if(	@{ $state->{heaviweightest_files} } < $SIZE_OF_TOP_FILES
-			or $state->{heaviweightest_files}->[-1]->{weight} < $file_weight_spent
-		) {
-			@{ $state->{heaviweightest_files} } = sort {$b->{weight} <=> $a->{weight}}
-				(@{ $state->{heaviweightest_files} }, {weight => $file_weight_spent, path => $task->[0]});
-			splice @{ $state->{heaviweightest_files} }, $SIZE_OF_TOP_FILES;
-		}
+	$state->{heaviweightest_files} ||= [];
+	if(@{ $state->{heaviweightest_files} } < 5 or $state->{heaviweightest_files}->[-1]->{weight} < $file_weight_spent) {
+		@{ $state->{heaviweightest_files} } = sort {$b->{weight} <=> $a->{weight}} (@{ $state->{heaviweightest_files} }, {weight => $file_weight_spent, path => $task->[0]});
+		splice @{ $state->{heaviweightest_files} }, 5;
 	}
+
 	
 	return @next;
 }
@@ -658,15 +539,14 @@ sub _free_up_space {
 	my($backups, $files, $blocks, $parts) = @{ $state->{db} }{qw(backups files blocks parts)};
 	
 	my $deleted = 0;
-	while(1) {
-		my($block_id, @files) = _get_block_to_delete($state);
-		last if ! $block_id;
-		next if exists $protected_block_ids->{ $block_id };
+	while ( @{ $state->{blocks_stack2delete} } ) {
+		my($block_id, @files) = @{ shift @{ $state->{blocks_stack2delete} } };
+		next if exists $protected_block_ids->{$block_id};
 		my $block = $blocks->find_row({ id => $block_id });
 		next if ! $block;
 		next if $block->{last_backup_id} == $state->{last_backup_id};
 		
-		$deleted += App::SimpleBackuper::_BlockDelete($options, $state, $block, $state->{blocks_info}->{ $block_id }->[2]);
+		$deleted += App::SimpleBackuper::_BlockDelete($options, $state, $block, \@files);
 		last if $deleted;
 	}
 	
